/**
 * Copyright (c) 2016-2018 Thales Global Services S.A.S.
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *   Abel Hegedus, Tamas Borbas, Balazs Grill, Peter Lunk, Daniel Segesdi (IncQuery Labs Ltd.) - initial API and implementation
 */
package org.eclipse.emf.diffmerge.patch.runtime;

import com.google.common.base.Objects;
import com.google.common.base.Optional;
import com.google.common.collect.Iterables;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Set;
import java.util.function.Consumer;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.diffmerge.api.IComparison;
import org.eclipse.emf.diffmerge.api.Role;
import org.eclipse.emf.diffmerge.api.diff.IDifference;
import org.eclipse.emf.diffmerge.api.diff.IPresenceDifference;
import org.eclipse.emf.diffmerge.api.diff.IReferenceValuePresence;
import org.eclipse.emf.diffmerge.diffdata.EAttributeValuePresence;
import org.eclipse.emf.diffmerge.diffdata.EElementPresence;
import org.eclipse.emf.diffmerge.diffdata.EMatch;
import org.eclipse.emf.diffmerge.diffdata.EReferenceValuePresence;
import org.eclipse.emf.diffmerge.patch.AttributeEntry;
import org.eclipse.emf.diffmerge.patch.ElementEntry;
import org.eclipse.emf.diffmerge.patch.ReferenceEntry;
import org.eclipse.emf.diffmerge.patch.api.ChangeDirection;
import org.eclipse.emf.diffmerge.patch.api.Identifiable;
import org.eclipse.emf.diffmerge.patch.api.ModelPatch;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchBuilder;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchEntry;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchEntryBuilder;
import org.eclipse.emf.diffmerge.patch.api.ModelPatchException;
import org.eclipse.emf.diffmerge.patch.runtime.identifier.EMFIdentifierProvider;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EDataType;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.xtend.lib.annotations.Accessors;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Conversions;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.ListExtensions;
import org.eclipse.xtext.xbase.lib.ObjectExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.eclipse.xtext.xbase.lib.Pure;

@SuppressWarnings("all")
public class ModelPatchRecorder {
  protected static class DiffComparator implements Comparator<IDifference> {
    public static final int ADD_ATTRIBUTE_PRIORITY = 7;
    
    public static final int ADD_REFERENCE_PRIORITY = 6;
    
    public static final int ADD_ELEMENT_PRIORITY = 5;
    
    public static final int ADD_CONTAINMENT_PRIORITY = 4;
    
    public static final int REMOVE_CONTAINMENT_PRIORITY = 3;
    
    public static final int REMOVE_ELEMENT_PRIORITY = 2;
    
    public static final int REMOVE_REFERENCE_PRIORITY = 1;
    
    public static final int REMOVE_ATTRIBUTE_PRIORITY = 0;
    
    public static ModelPatchRecorder.DiffComparator getINSTANCE() {
      return new ModelPatchRecorder.DiffComparator();
    }
    
    @Override
    public int compare(final IDifference o1, final IDifference o2) {
      int _priority = this.getPriority(o1);
      int _priority_1 = this.getPriority(o2);
      final int result = (_priority - _priority_1);
      if ((result != 0)) {
        return result;
      }
      return this.compareTo(o1, o2);
    }
    
    protected int _compareTo(final IDifference o1, final IDifference o2) {
      return 0;
    }
    
    protected int _compareTo(final EElementPresence o1, final EElementPresence o2) {
      int _priority = this.getPriority(o1);
      boolean _equals = (_priority == ModelPatchRecorder.DiffComparator.ADD_ELEMENT_PRIORITY);
      if (_equals) {
        int _deepness = this.getDeepness(o1.getElement());
        int _deepness_1 = this.getDeepness(o2.getElement());
        return (_deepness - _deepness_1);
      } else {
        int _deepness_2 = this.getDeepness(o2.getElement());
        int _deepness_3 = this.getDeepness(o1.getElement());
        return (_deepness_2 - _deepness_3);
      }
    }
    
    protected int _compareTo(final EReferenceValuePresence o1, final EReferenceValuePresence o2) {
      int _priority = this.getPriority(o1);
      boolean _equals = (_priority == ModelPatchRecorder.DiffComparator.ADD_CONTAINMENT_PRIORITY);
      if (_equals) {
        int _deepness = this.getDeepness(o1.getValueMatch().getTarget());
        int _deepness_1 = this.getDeepness(o2.getValueMatch().getTarget());
        return (_deepness - _deepness_1);
      } else {
        int _priority_1 = this.getPriority(o1);
        boolean _equals_1 = (_priority_1 == ModelPatchRecorder.DiffComparator.REMOVE_CONTAINMENT_PRIORITY);
        if (_equals_1) {
          int _deepness_2 = this.getDeepness(o2.getValueMatch().getReference());
          int _deepness_3 = this.getDeepness(o1.getValueMatch().getReference());
          return (_deepness_2 - _deepness_3);
        }
      }
      return 0;
    }
    
    private int getDeepness(final EObject eObject) {
      EObject _eContainer = eObject.eContainer();
      boolean _tripleEquals = (_eContainer == null);
      if (_tripleEquals) {
        return 0;
      }
      int _deepness = this.getDeepness(eObject.eContainer());
      return (_deepness + 1);
    }
    
    protected int _getPriority(final IDifference diff) {
      try {
        throw new ModelPatchException("Unsupported difference type");
      } catch (Throwable _e) {
        throw Exceptions.sneakyThrow(_e);
      }
    }
    
    protected int _getPriority(final EAttributeValuePresence diff) {
      Role _presenceRole = diff.getPresenceRole();
      boolean _equals = Objects.equal(Role.TARGET, _presenceRole);
      if (_equals) {
        return ModelPatchRecorder.DiffComparator.ADD_ATTRIBUTE_PRIORITY;
      }
      return ModelPatchRecorder.DiffComparator.REMOVE_ATTRIBUTE_PRIORITY;
    }
    
    protected int _getPriority(final EElementPresence diff) {
      Role _presenceRole = diff.getPresenceRole();
      boolean _equals = Objects.equal(Role.TARGET, _presenceRole);
      if (_equals) {
        return ModelPatchRecorder.DiffComparator.ADD_ELEMENT_PRIORITY;
      }
      return ModelPatchRecorder.DiffComparator.REMOVE_ELEMENT_PRIORITY;
    }
    
    protected int _getPriority(final EReferenceValuePresence diff) {
      Role _presenceRole = diff.getPresenceRole();
      boolean _equals = Objects.equal(Role.TARGET, _presenceRole);
      if (_equals) {
        EStructuralFeature _feature = diff.getFeature();
        boolean _isContainment = ((EReference) _feature).isContainment();
        if (_isContainment) {
          return ModelPatchRecorder.DiffComparator.ADD_CONTAINMENT_PRIORITY;
        }
        return ModelPatchRecorder.DiffComparator.ADD_REFERENCE_PRIORITY;
      }
      EStructuralFeature _feature_1 = diff.getFeature();
      boolean _isContainment_1 = ((EReference) _feature_1).isContainment();
      if (_isContainment_1) {
        return ModelPatchRecorder.DiffComparator.REMOVE_CONTAINMENT_PRIORITY;
      }
      return ModelPatchRecorder.DiffComparator.REMOVE_REFERENCE_PRIORITY;
    }
    
    public int compareTo(final IDifference o1, final IDifference o2) {
      if (o1 instanceof EReferenceValuePresence
           && o2 instanceof EReferenceValuePresence) {
        return _compareTo((EReferenceValuePresence)o1, (EReferenceValuePresence)o2);
      } else if (o1 instanceof EElementPresence
           && o2 instanceof EElementPresence) {
        return _compareTo((EElementPresence)o1, (EElementPresence)o2);
      } else if (o1 != null
           && o2 != null) {
        return _compareTo(o1, o2);
      } else {
        throw new IllegalArgumentException("Unhandled parameter types: " +
          Arrays.<Object>asList(o1, o2).toString());
      }
    }
    
    public int getPriority(final IDifference diff) {
      if (diff instanceof EAttributeValuePresence) {
        return _getPriority((EAttributeValuePresence)diff);
      } else if (diff instanceof EReferenceValuePresence) {
        return _getPriority((EReferenceValuePresence)diff);
      } else if (diff instanceof EElementPresence) {
        return _getPriority((EElementPresence)diff);
      } else if (diff != null) {
        return _getPriority(diff);
      } else {
        throw new IllegalArgumentException("Unhandled parameter types: " +
          Arrays.<Object>asList(diff).toString());
      }
    }
  }
  
  @Accessors
  @Extension
  private EMFIdentifierProvider identifierProvider = new EMFIdentifierProvider();
  
  private ModelPatchBuilder attributeRemoveBuilder;
  
  private ModelPatchBuilder referenceRemoveBuilder;
  
  private ModelPatchBuilder containmentRemoveBuilder;
  
  private ModelPatchBuilder secondContainmentRemoveBuilder;
  
  private ModelPatchBuilder elementRemoveBuilder;
  
  private ModelPatchBuilder elementAddBuilder;
  
  private ModelPatchBuilder containmentAddBuilder;
  
  private ModelPatchBuilder referenceAddBuilder;
  
  private ModelPatchBuilder attributeAddBuilder;
  
  private ModelPatchBuilder modelPatchBuilder;
  
  private Set<ModelPatchEntry> addedEntries = CollectionLiterals.<ModelPatchEntry>newHashSet();
  
  public ModelPatch generateModelPatch(final IComparison comparison) {
    ArrayList<IDifference> _newArrayList = CollectionLiterals.<IDifference>newArrayList();
    final Procedure1<ArrayList<IDifference>> _function = (ArrayList<IDifference> it) -> {
      List<IDifference> _differences = comparison.getDifferences(Role.REFERENCE);
      Iterables.<IDifference>addAll(it, _differences);
      List<IDifference> _differences_1 = comparison.getDifferences(Role.TARGET);
      Iterables.<IDifference>addAll(it, _differences_1);
    };
    final ArrayList<IDifference> differences = ObjectExtensions.<ArrayList<IDifference>>operator_doubleArrow(_newArrayList, _function);
    return this.generateModelPatch(differences);
  }
  
  public ModelPatch generateModelPatch(final Iterable<? extends IDifference> diffs) {
    final List<IDifference> differences = ListExtensions.<IDifference>sortInplace(CollectionLiterals.<IDifference>newArrayList(((IDifference[])Conversions.unwrapArray(diffs, IDifference.class))), ModelPatchRecorder.DiffComparator.getINSTANCE());
    this.initModelPatchBuilders();
    final Consumer<IDifference> _function = (IDifference it) -> {
      this.processDifference(it);
    };
    differences.forEach(_function);
    final ModelPatch modelPatch = this.appendNewEntries(this.appendNewEntries(this.appendNewEntries(this.appendNewEntries(this.appendNewEntries(this.appendNewEntries(this.appendNewEntries(this.appendNewEntries(this.appendNewEntries(this.modelPatchBuilder, this.attributeRemoveBuilder), this.referenceRemoveBuilder), this.containmentRemoveBuilder), this.secondContainmentRemoveBuilder), this.elementRemoveBuilder), this.elementAddBuilder), this.containmentAddBuilder), this.referenceAddBuilder), this.attributeAddBuilder).build();
    return modelPatch;
  }
  
  protected ModelPatchBuilder initModelPatchBuilders() {
    ModelPatchBuilder _xblockexpression = null;
    {
      this.attributeRemoveBuilder = ModelPatchBuilder.create();
      this.referenceRemoveBuilder = ModelPatchBuilder.create();
      this.containmentRemoveBuilder = ModelPatchBuilder.create();
      this.secondContainmentRemoveBuilder = ModelPatchBuilder.create();
      this.elementRemoveBuilder = ModelPatchBuilder.create();
      this.elementAddBuilder = ModelPatchBuilder.create();
      this.containmentAddBuilder = ModelPatchBuilder.create();
      this.referenceAddBuilder = ModelPatchBuilder.create();
      this.attributeAddBuilder = ModelPatchBuilder.create();
      _xblockexpression = this.modelPatchBuilder = ModelPatchBuilder.create();
    }
    return _xblockexpression;
  }
  
  private void processDifference(final IDifference diff) {
    if ((diff instanceof IPresenceDifference)) {
      final Role mergeDestination = Role.REFERENCE;
      final Role presenceRole = ((IPresenceDifference)diff).getPresenceRole();
      final ChangeDirection direction = this.getChangeDirection(presenceRole);
      this.processDifference(diff, mergeDestination, presenceRole, direction);
    }
  }
  
  private void _processDifference(final IDifference diff, final Role mergeDestination, final Role presenceRole, final ChangeDirection direction) {
    try {
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("Unsupported IDifference descendant: ");
      String _name = diff.getClass().getName();
      _builder.append(_name);
      throw new ModelPatchException(_builder.toString());
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private void _processDifference(final EAttributeValuePresence diff, final Role mergeDestination, final Role presenceRole, final ChangeDirection direction) {
    final EObject owner = diff.getElementMatch().get(mergeDestination);
    final EStructuralFeature feature = diff.getFeature();
    final Object value = diff.getValue();
    final EDataType eAttributeType = ((EAttribute) feature).getEAttributeType();
    String _id = this.getId(owner);
    final Identifiable ownerId = new Identifiable(_id);
    String _identify = this.identifierProvider.identify(feature);
    final Identifiable featureId = new Identifiable(_identify);
    final String stringValue = EcoreUtil.convertToString(eAttributeType, value);
    final EObject presenceOwner = diff.getElementMatch().get(presenceRole);
    final Optional<Integer> index = this.indexOf(presenceOwner, value, feature);
    ModelPatchEntryBuilder _entryBuilder = ModelPatchBuilder.entryBuilder(ownerId, direction);
    final Procedure1<ModelPatchEntryBuilder> _function = (ModelPatchEntryBuilder it) -> {
      it.setFeature(featureId);
      it.setValue(stringValue);
      it.setIndex(index);
    };
    final ModelPatchEntryBuilder entryBuilder = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_entryBuilder, _function);
    final AttributeEntry attributeEntry = entryBuilder.buildAttributeEntry();
    boolean _equals = Objects.equal(direction, ChangeDirection.ADD);
    if (_equals) {
      boolean _isMany = feature.isMany();
      boolean _not = (!_isMany);
      if (_not) {
        final Object oldValue = owner.eGet(feature);
        if ((oldValue != null)) {
          final String oldStringValue = EcoreUtil.convertToString(eAttributeType, oldValue);
          final Procedure1<ModelPatchEntryBuilder> _function_1 = (ModelPatchEntryBuilder it) -> {
            it.setDirection(ChangeDirection.REMOVE);
            it.setValue(oldStringValue);
          };
          final AttributeEntry removeEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_1).buildAttributeEntry();
          this.addNewEntry(this.attributeRemoveBuilder, removeEntry);
        }
      }
      this.addNewEntry(this.attributeAddBuilder, attributeEntry);
    } else {
      this.addNewEntry(this.attributeRemoveBuilder, attributeEntry);
    }
  }
  
  private Optional<Integer> indexOf(final EObject presenceOwner, final Object value, final EStructuralFeature feature) {
    boolean _isMany = feature.isMany();
    if (_isMany) {
      Object _eGet = presenceOwner.eGet(feature);
      final EList values = ((EList) _eGet);
      final int index = values.indexOf(value);
      if ((index < 0)) {
        throw new IllegalStateException("The object is not found in the containing list.");
      } else {
        return Optional.<Integer>of(Integer.valueOf(index));
      }
    }
    return Optional.<Integer>absent();
  }
  
  private void _processDifference(final EReferenceValuePresence diff, final Role mergeDestination, final Role presenceRole, final ChangeDirection direction) {
    final EObject owner = diff.getElementMatch().get(mergeDestination);
    EStructuralFeature _feature = diff.getFeature();
    final EReference feature = ((EReference) _feature);
    EObject _xifexpression = null;
    EMatch _valueMatch = diff.getValueMatch();
    boolean _tripleNotEquals = (_valueMatch != null);
    if (_tripleNotEquals) {
      _xifexpression = diff.getValueMatch().get(presenceRole);
    } else {
      _xifexpression = diff.getValue();
    }
    final EObject value = _xifexpression;
    String _id = this.getId(owner);
    final Identifiable ownerId = new Identifiable(_id);
    String _identify = this.identifierProvider.identify(feature);
    final Identifiable featureId = new Identifiable(_identify);
    String _id_1 = this.getId(value);
    final Identifiable targetId = new Identifiable(_id_1);
    final EObject presenceOwner = diff.getElementMatch().get(presenceRole);
    final Optional<Integer> index = this.indexOf(presenceOwner, value, feature);
    ModelPatchEntryBuilder _entryBuilder = ModelPatchBuilder.entryBuilder(ownerId, direction);
    final Procedure1<ModelPatchEntryBuilder> _function = (ModelPatchEntryBuilder it) -> {
      it.setFeature(featureId);
      it.setTarget(targetId);
      it.setIndex(index);
    };
    final ModelPatchEntryBuilder entryBuilder = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_entryBuilder, _function);
    final ReferenceEntry referenceEntry = entryBuilder.buildReferenceEntry();
    boolean _equals = Objects.equal(direction, ChangeDirection.ADD);
    if (_equals) {
      boolean _isMany = feature.isMany();
      boolean _not = (!_isMany);
      if (_not) {
        final Object oldValue = owner.eGet(feature);
        if ((oldValue != null)) {
          if ((oldValue instanceof EObject)) {
            String _id_2 = this.getId(((EObject)oldValue));
            final Identifiable oldValueId = new Identifiable(_id_2);
            final Procedure1<ModelPatchEntryBuilder> _function_1 = (ModelPatchEntryBuilder it) -> {
              it.setDirection(ChangeDirection.REMOVE);
              it.setTarget(oldValueId);
            };
            final ReferenceEntry removeEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_1).buildReferenceEntry();
            boolean _isContainment = feature.isContainment();
            if (_isContainment) {
              this.addNewEntry(this.secondContainmentRemoveBuilder, removeEntry);
            } else {
              this.addNewEntry(this.referenceRemoveBuilder, removeEntry);
            }
          }
        }
      }
      boolean _isContainment_1 = feature.isContainment();
      if (_isContainment_1) {
        this.addNewEntry(this.containmentAddBuilder, referenceEntry);
      } else {
        this.addNewEntry(this.referenceAddBuilder, referenceEntry);
      }
    } else {
      boolean _isContainment_2 = feature.isContainment();
      if (_isContainment_2) {
        this.addNewEntry(this.secondContainmentRemoveBuilder, referenceEntry);
      } else {
        this.addNewEntry(this.referenceRemoveBuilder, referenceEntry);
      }
    }
  }
  
  private void _processDifference(final EElementPresence diff, final Role mergeDestination, final Role presenceRole, final ChangeDirection direction) {
    final EObject element = diff.getElement();
    EObject parent = diff.getOwnerMatch().get(mergeDestination);
    final IReferenceValuePresence ownershipDifference = diff.getElementMatch().getOwnershipDifference(presenceRole);
    final boolean containmentHasPresence = (ownershipDifference != null);
    EReference containmentFeature = null;
    if ((ownershipDifference != null)) {
      containmentFeature = ownershipDifference.getFeature();
    } else {
      EStructuralFeature _eContainingFeature = element.eContainingFeature();
      containmentFeature = ((EReference) _eContainingFeature);
      parent = element.eContainer();
    }
    String _id = this.getId(element);
    final Identifiable elementId = new Identifiable(_id);
    String _identify = this.identifierProvider.identify(element.eClass());
    final Identifiable typeId = new Identifiable(_identify);
    String _id_1 = this.getId(parent);
    final Identifiable parentId = new Identifiable(_id_1);
    String _identify_1 = this.identifierProvider.identify(containmentFeature);
    final Identifiable containmentFeatureId = new Identifiable(_identify_1);
    final Optional<Integer> index = this.indexOf(element.eContainer(), element, containmentFeature);
    boolean _equals = Objects.equal(direction, ChangeDirection.ADD);
    if (_equals) {
      ModelPatchEntryBuilder _entryBuilder = ModelPatchBuilder.entryBuilder(elementId, direction);
      final Procedure1<ModelPatchEntryBuilder> _function = (ModelPatchEntryBuilder it) -> {
        it.setType(typeId);
      };
      final ElementEntry elementEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_entryBuilder, _function).buildElementEntry();
      this.addNewEntry(this.elementAddBuilder, elementEntry);
      ModelPatchEntryBuilder _entryBuilder_1 = ModelPatchBuilder.entryBuilder(parentId, direction);
      final Procedure1<ModelPatchEntryBuilder> _function_1 = (ModelPatchEntryBuilder it) -> {
        it.setFeature(containmentFeatureId);
        it.setTarget(elementId);
        it.setIndex(index);
      };
      final ModelPatchEntryBuilder entryBuilder = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_entryBuilder_1, _function_1);
      final ReferenceEntry referenceEntry = entryBuilder.buildReferenceEntry();
      boolean _isMany = containmentFeature.isMany();
      boolean _not = (!_isMany);
      if (_not) {
        final Object oldValue = parent.eGet(containmentFeature);
        if (((oldValue != null) && (!Objects.equal(oldValue, element)))) {
          if ((oldValue instanceof EObject)) {
            String _id_2 = this.getId(((EObject)oldValue));
            final Identifiable oldValueId = new Identifiable(_id_2);
            final Procedure1<ModelPatchEntryBuilder> _function_2 = (ModelPatchEntryBuilder it) -> {
              it.setDirection(ChangeDirection.REMOVE);
              it.setTarget(oldValueId);
            };
            final ReferenceEntry removeEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_2).buildReferenceEntry();
            this.addNewEntry(this.containmentRemoveBuilder, removeEntry);
          }
        }
      }
      if ((!containmentHasPresence)) {
        this.addNewEntry(this.containmentAddBuilder, referenceEntry);
      }
      final Procedure1<ModelPatchEntryBuilder> _function_3 = (ModelPatchEntryBuilder it) -> {
        it.setDirection(ChangeDirection.ADD);
        it.setContext(elementId);
        it.setIndex(Optional.<Integer>absent());
      };
      ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_3);
      this.saveAllAttribute(element, this.attributeAddBuilder, entryBuilder);
      this.saveAllNonContainmentRef(element, this.referenceAddBuilder, entryBuilder);
    } else {
      final ModelPatchEntryBuilder elementAsContextEntryBuilder = ModelPatchBuilder.entryBuilder(elementId, direction);
      this.saveAllAttribute(element, this.attributeRemoveBuilder, elementAsContextEntryBuilder);
      this.saveAllNonContainmentRef(element, this.referenceRemoveBuilder, elementAsContextEntryBuilder);
      if ((!containmentHasPresence)) {
        ModelPatchEntryBuilder _entryBuilder_2 = ModelPatchBuilder.entryBuilder(parentId, direction);
        final Procedure1<ModelPatchEntryBuilder> _function_4 = (ModelPatchEntryBuilder it) -> {
          it.setFeature(containmentFeatureId);
          it.setTarget(elementId);
          it.setIndex(index);
        };
        final ReferenceEntry removeableContainmentEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(_entryBuilder_2, _function_4).buildReferenceEntry();
        this.addNewEntry(this.containmentRemoveBuilder, removeableContainmentEntry);
      }
      final Procedure1<ModelPatchEntryBuilder> _function_5 = (ModelPatchEntryBuilder it) -> {
        it.setType(typeId);
      };
      final ElementEntry elementRemove = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(elementAsContextEntryBuilder, _function_5).buildElementEntry();
      this.addNewEntry(this.elementRemoveBuilder, elementRemove);
    }
  }
  
  private void saveAllAttribute(final EObject element, final ModelPatchBuilder patchBuilder, final ModelPatchEntryBuilder entryBuilder) {
    final Function1<EAttribute, Boolean> _function = (EAttribute it) -> {
      boolean _isID = it.isID();
      return Boolean.valueOf((!_isID));
    };
    Iterable<EAttribute> _filter = IterableExtensions.<EAttribute>filter(element.eClass().getEAllAttributes(), _function);
    for (final EAttribute eAttr : _filter) {
      boolean _eIsSet = element.eIsSet(eAttr);
      if (_eIsSet) {
        final Object value = element.eGet(eAttr);
        if ((value instanceof EList)) {
          for (final Object v : ((EList)value)) {
            {
              final Optional<Integer> index = this.indexOf(element, v, eAttr);
              final Procedure1<ModelPatchEntryBuilder> _function_1 = (ModelPatchEntryBuilder it) -> {
                String _identify = this.identifierProvider.identify(eAttr);
                Identifiable _identifiable = new Identifiable(_identify);
                it.setFeature(_identifiable);
                it.setValue(EcoreUtil.convertToString(eAttr.getEAttributeType(), v));
                it.setIndex(index);
              };
              final AttributeEntry addedAttributeEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_1).buildAttributeEntry();
              this.addNewEntry(patchBuilder, addedAttributeEntry);
            }
          }
        } else {
          final Optional<Integer> index = this.indexOf(element, value, eAttr);
          final Procedure1<ModelPatchEntryBuilder> _function_1 = (ModelPatchEntryBuilder it) -> {
            String _identify = this.identifierProvider.identify(eAttr);
            Identifiable _identifiable = new Identifiable(_identify);
            it.setFeature(_identifiable);
            it.setValue(EcoreUtil.convertToString(eAttr.getEAttributeType(), value));
            it.setIndex(index);
          };
          final AttributeEntry addedAttributeEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_1).buildAttributeEntry();
          this.addNewEntry(patchBuilder, addedAttributeEntry);
        }
      }
    }
  }
  
  private void saveAllNonContainmentRef(final EObject element, final ModelPatchBuilder patchBuilder, final ModelPatchEntryBuilder entryBuilder) {
    final Function1<EReference, Boolean> _function = (EReference it) -> {
      return Boolean.valueOf(((((!it.isContainment()) && (!it.isDerived())) && it.isChangeable()) && (((it.getEOpposite() == null) || it.getEOpposite().isDerived()) || (!it.getEOpposite().isContainment()))));
    };
    Iterable<EReference> _filter = IterableExtensions.<EReference>filter(element.eClass().getEAllReferences(), _function);
    for (final EReference eRef : _filter) {
      {
        final Object value = element.eGet(eRef);
        if ((value != null)) {
          if ((value instanceof EObject)) {
            final Optional<Integer> index = this.indexOf(element, value, eRef);
            final Procedure1<ModelPatchEntryBuilder> _function_1 = (ModelPatchEntryBuilder it) -> {
              String _identify = this.identifierProvider.identify(eRef);
              Identifiable _identifiable = new Identifiable(_identify);
              it.setFeature(_identifiable);
              String _id = this.getId(((EObject)value));
              Identifiable _identifiable_1 = new Identifiable(_id);
              it.setTarget(_identifiable_1);
              it.setIndex(index);
            };
            final ReferenceEntry removeableReferenceEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_1).buildReferenceEntry();
            this.addNewEntry(patchBuilder, removeableReferenceEntry);
          } else {
            if ((value instanceof EList)) {
              for (final Object v : ((EList)value)) {
                if ((v instanceof EObject)) {
                  final Optional<Integer> index_1 = this.indexOf(element, v, eRef);
                  final Procedure1<ModelPatchEntryBuilder> _function_2 = (ModelPatchEntryBuilder it) -> {
                    String _identify = this.identifierProvider.identify(eRef);
                    Identifiable _identifiable = new Identifiable(_identify);
                    it.setFeature(_identifiable);
                    String _id = this.getId(((EObject)v));
                    Identifiable _identifiable_1 = new Identifiable(_id);
                    it.setTarget(_identifiable_1);
                    it.setIndex(index_1);
                  };
                  final ReferenceEntry referenceEntry = ObjectExtensions.<ModelPatchEntryBuilder>operator_doubleArrow(entryBuilder, _function_2).buildReferenceEntry();
                  this.addNewEntry(patchBuilder, referenceEntry);
                }
              }
            }
          }
        }
      }
    }
  }
  
  private ModelPatchBuilder addNewEntry(final ModelPatchBuilder builder, final ModelPatchEntry entry) {
    boolean _contains = this.addedEntries.contains(entry);
    boolean _not = (!_contains);
    if (_not) {
      this.addedEntries.add(entry);
      return builder.addEntry(entry);
    }
    return builder;
  }
  
  private ModelPatchBuilder appendNewEntries(final ModelPatchBuilder target, final ModelPatchBuilder source) {
    target.append(source);
    return target;
  }
  
  private ChangeDirection getChangeDirection(final Role scopeDiff) {
    boolean _equals = Objects.equal(scopeDiff, Role.TARGET);
    if (_equals) {
      return ChangeDirection.ADD;
    }
    return ChangeDirection.REMOVE;
  }
  
  private String getId(final EObject eObject) {
    return this.identifierProvider.identifyEObject(eObject);
  }
  
  private void processDifference(final IDifference diff, final Role mergeDestination, final Role presenceRole, final ChangeDirection direction) {
    if (diff instanceof EAttributeValuePresence) {
      _processDifference((EAttributeValuePresence)diff, mergeDestination, presenceRole, direction);
      return;
    } else if (diff instanceof EReferenceValuePresence) {
      _processDifference((EReferenceValuePresence)diff, mergeDestination, presenceRole, direction);
      return;
    } else if (diff instanceof EElementPresence) {
      _processDifference((EElementPresence)diff, mergeDestination, presenceRole, direction);
      return;
    } else if (diff != null) {
      _processDifference(diff, mergeDestination, presenceRole, direction);
      return;
    } else {
      throw new IllegalArgumentException("Unhandled parameter types: " +
        Arrays.<Object>asList(diff, mergeDestination, presenceRole, direction).toString());
    }
  }
  
  @Pure
  public EMFIdentifierProvider getIdentifierProvider() {
    return this.identifierProvider;
  }
  
  public void setIdentifierProvider(final EMFIdentifierProvider identifierProvider) {
    this.identifierProvider = identifierProvider;
  }
}
